/*
  * Originates in the work for a seminar in the winter semester 2000/2001
  * at the University of Essen (http://www.uni-essen.de), department:
  * "MIS and software development" (http://nestroy.wi-inf.uni-essen.de).
  *
  * This DLL registers the function "BSF" with Rexx;
  * it allows Java to call into Rexx using the fine RexxSAA-interface
  * conventions and JNI ("Java Native Interface").
  *
  * version:
  *                2.0.0 2003-04-16, final version to be introduced at the 2003 International Rexx Symposium
  *                                  caters for IBM (BSF 2.2) and Apache (BSF 2.3 and higher)
  *
  *                1.9.4 2003-04-14, adjusted Rexx memory to what it is supposed to be, ie.:
  *                                  - never release memory allocated by Rexx
  *                                  - if allocating memory, then freeing it is o.k.
  *                                  - Rexx creates RXSTRINGS with a buffer size of 256L (RXAUTOBUFLEN)
  *                                    if size is not sufficient, create a larger buffer, but do
  *                                    NOT free Rexx'
  *
  *                1.9.3 2003-04-05, adjusted for Linux (RH 7.2)
  *                1.9.2 2003-02-19, switched arguments for BsfLoadJava, now arg(1)=classpath,
  *                     arg(2)=Append/Prepend;
  *                1.9.1 2003-02-12, adjusting for JNI 1.1 *and* 1.2 (as starting with
  *                     Java 1.4 the JNI 1.1 interface does not work anymore (just bombs)
  *
  *                1.9.0 2003-01-27, ---rgf, added ability to load JVM from the Rexx side,
  *                     initializing BSF4Rexx from the Java side as well, such that all of
  *                     BSF4Rexx' functionality is available to Rexx programs started by Rexx
  *
  *                1.2.1 2003-01-06, ---rgf, finally found bug causing Object Rexx to bomb (but also Regina, it turned out!)
  *                               reason: freeing memory for Rexx-code with wrong method !
  *
  *                2003-01-05, ---rgf, adapted to Mark Hessling's great "Rexx/Trans"
  *                      (cf. <http://rexxtrans.sourceforge.net/>), which allows for
  *                      generic support for any supported Rexx interpreter, e.g.:
  *
  *                         Regina <http://regina-rexx.sourceforge.net/>,
  *                         IBM Object Rexx <http://www.software.ibm.com/ad/obj-rexx/>,
  *                         Quercus Systems Personal Rexx <http://www.quercus-sys.com/>,
  *                         Enterprise Rexx <http://www.WinREXX.com/>
  *
  *                 2001-06-11, ---rgf, runs with JNI 1 (does *not* need JRE 1.2 or higher
  *                         anymore, so OS/2 users with 1.1.8 can use this package as well)
  *
  * author:  (c) 2001-2003 Rony G. Flatscher, WU (pronounced: "vey-u") Vienna University of Economics
  *                          and Business Administration
  *                       (preliminary work by Peter Kalender (Oct 2000 to Feb 2001), University of Essen)
  *
  *  LICENSE:
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">
  *              http://www.gnu.org/copyleft/lesser.html
  *     </a>

*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define INCL_RXSHV
#define INCL_RXFUNC

// ---------------------- linking to which Rexx ----
#ifdef USE_REXXTRANS
# include <rexxtrans.h>

#elif defined USE_OREXX
# include <rexx.h>

#else
# include <rexxsaa.h>
#endif

#ifdef _MSC_VER
# pragma warning(disable:4100)
#endif


#ifdef OS2
   #define INCL_DOSMEMMGR
   #include <os2.h>
   #include <bsememf.h>
#endif

/* ---> MH commented it out, now it's time to even delete it
#if 0
        // Java support
#include <jni.h>                        // "Java Native Interface" (JNI)
#endif
*/


        // generated by Java to support the JNI
        // see comments at the end of "RexxAndJava.java"
#ifdef OS2
# if BSF4REXX_JDK != 118
#  define PROTOTYPE_JNI_ONLY
#    ifdef APACHE
#       include "org_apache_bsf_engines_rexx_RexxAndJava.h"
#    else
#       include "com_ibm_bsf_engines_rexx_RexxAndJava.h"
#    endif
# else
#    ifdef APACHE
#       include "org_apache_bsf_engines_rexx_RexxAndJava.h"
#    else
#       include "com_ibm_bsf_engines_rexx_RexxAndJava.h"
#    endif
# endif
#else
#    ifdef APACHE
#       include "org_apache_bsf_engines_rexx_RexxAndJava.h"
#    else
#       include "com_ibm_bsf_engines_rexx_RexxAndJava.h"
#    endif
#endif

#ifndef JNI_VERSION_1_1
    #define JNI_VERSION_1_1 0x00010001
#endif

      // under Linux gcc RH 7.2 not defined, ---rgf, 2003-04-05
#ifndef FALSE
    #define FALSE 0
#endif

#ifndef INT
    #define INT int
#endif
    // end (Linux RH 7.2)



#ifdef __cplusplus
extern "C" {            // make sure no C++ name mangling comes into the way
#endif


#ifdef APACHE
#   define BSF_VERSION     "200.20030416 org/apache/bsf/engines/rexx"  // version: "MajorNumber"."YYYYMMDD"
#   define JAVA_4_REXX     "org/apache/bsf/engines/rexx/Java4Rexx"  // 2003-04-16, supportive Java4Rexx-class
#else
#   define BSF_VERSION     "200.20030416 com/ibm/bsf/engines/rexx"  // version: "MajorNumber"."YYYYMMDD"
#   define JAVA_4_REXX     "com/ibm/bsf/engines/rexx/Java4Rexx"    // 2003-01-20, supportive Java4Rexx-class
#endif

#define DLLNAME         "BSF4Rexx"      // name of DLL and name for "PARSE SOURCE"
#define JAVA_CALL_BSF   "javaCallBSF"   // name of Java method to carry out the BSF-call

#define  INVALID_ROUTINE 40            /* Rexx: raise Rexx error # 40   */
#define  VALID_ROUTINE    0            /* Rexx: successful completion   */

JNIEnv    *oldGlobEnv =NULL, *globEnv =NULL;          // used by BSF() to call back to Java
jobject    oldGlobObj =NULL,  globObj =NULL;          // used by BSF() to call back to Java
jmethodID  oldGlobMid =NULL,  globMid =NULL;          // used by BSF() to call back to Java
JavaVM    *oldGlobJvm =NULL, *globJvm =NULL;          // used by BsfLoadJava(), BsfUnloadJava()
jobject    oldGlobORef=NULL,  globORef=NULL;          // this stores the global reference to the object, pinning the JVM

/*
static CRITICAL_SECTION bsfLock;    // 2003-01-23, ---rgf, try to synchronize, if
                                    // ORexx programs call BSF() from multiple threads
*/


#ifdef UNIX
#define PATH_SEPARATOR ':'
#else   /* default to: DOS, WIN[32], OS2 */
#define PATH_SEPARATOR ';'
#endif


inline void zeigeGlobs()    // show values for global variables
{
/*
    fprintf(stderr, "\tglobEnv=%d, globObj=%d, globMid=%d,\n\tglobVm=%d, globORef=%d\n",
                    globEnv, globObj, globMid, globJvm, globORef);
*/
}

        // inline function for taking care of the different ways of freeing Rexx-allocated memory
inline void RgfFreeMemory (void * memory)
{

//    printf(">>> RgfFreeMemory() ... \n");
#if defined (USE_REGINA) || defined (USE_REXXTRANS)     // 2003-01-04, ---rgf, added defined (USE_REXXTRANS)
//    printf(">>> RgfFreeMemory(): 'USE_REGINA' or 'USE_REXXTRANS' defined, RexxFreeMemory()\n");
    RexxFreeMemory( memory );

#elif defined WIN32
//    printf(">>> RgfFreeMemory(): 'WIN32' defined, GlobalFree()\n");
    GlobalFree( memory );

#elif defined OS2
//    printf(">>> RgfFreeMemory(): 'OS2' defined, DosFreeMem()\n");
    DosFreeMem( memory );

#else
//    printf(">>> RgfFreeMemory():  using free()\n");
    free( memory );
#endif

}



    // inline function for reserving memory disposable for Rexx (guessing how Rexx is allocating it)
            // usage, e.g.: PSZ mem = (PSZ) RgfAllocateMemory(3451l);
inline PVOID RgfAllocateMemory(long size)
{
//    printf(">>> RgfAllocateMemory() ... \n");
#if defined (USE_REGINA) || defined (USE_REXXTRANS)     // 2003-01-04, ---rgf, added defined (USE_REXXTRANS)
//    printf(">>> RgfAllocateMemory(): 'USE_REGINA' or 'USE_REXXTRANS' defined, RexxFreeMemory()\n");
    return RexxAllocateMemory( size );       // RexxAllocMemory( nrOfBytes );

#elif defined WIN32
//    printf(">>> RgfAllocateMemory(): 'WIN32' defined, GlobalFree()\n");
    // #define GMEM_FIXED 0x0000       // according to docs: if GMEM_FIXED is set, then a pointer to memory is returned
    return GlobalLock( GlobalAlloc( GMEM_FIXED, size) );

#elif defined OS2
//    printf(">>> RgfAllocateMemory(): 'OS2' defined, DosFreeMem()\n");
    PVOID pmem = NULL;              // define a typeless pointer
    // #define PAG_WRITE 0x0000002     // define writable (should imply PAG_READ which would be 0x0000 0001)
    DosAllocMem( &pmem, size, PAG_WRITE | PAG_COMMIT );  // PAG_COMMIT from RexxTrans 1.6 (didn't find fPerm)
    return pmem;

#else       // hmm, don't know whether this makes sense
//    printf(">>> RgfAllocateMemory():  using malloc()\n");
    return malloc( size );
#endif
}




/*  Save Java environment handle and java object handle, find for that environment and
 *  instance the method id for being able to call back into Java later via Rexx using
 *  the function BSF() in here. ---rgf, 2003-03-15
 */
inline void set_global_vars (JNIEnv * env, jobject obj)
{
    globEnv = env;  // assign present Java environment handle
    globObj = obj;  // assign present Java object handle

           // get Java callback method via the Java object's class, looking up a method
           // named according to JAVA_CALL_BSF, signature: one argument (a String array) and
           // returning a String
    globMid = (*globEnv).GetMethodID( (*globEnv).GetObjectClass(globObj),
                                  JAVA_CALL_BSF,
                                  "([Ljava/lang/String;)Ljava/lang/String;" );
}





   // declare functions (define signatures)
APIRET APIENTRY BSF                        (   PSZ name, ULONG argc, PRXSTRING argv, PSZ queuename, PRXSTRING retstr);

   // same entries, slightly different signature than above
APIRET APIENTRY BsfInvokedBy               ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfLoadJava                ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfUnloadJava              ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfLoadFuncs               ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfDropFuncs               ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfVersion                 ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfQueryAllFunctions        ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );
APIRET APIENTRY BsfQueryRegisteredFunctions ( PSZ name, LONG numargs, RXSTRING args[], PSZ queuename, PRXSTRING retstr );


/*********************************************************************/
/* ApiFncTable                                                       */
/*   Array of names of the REXXASP1 functions.                       */
/*   This list is used for registration and deregistration.          */
/*********************************************************************/
PSZ  ApiFncTable[] =     // table of functions to be registered with Rexx
   {
       "BSF",                       // interface for Rexx to Java
       "BsfInvokedBy",              // allows querying for invocation type of this DLL
       "BsfLoadJava",               // load the JVM by creating a new instance of BSFManager, if loaded by Rexx, using JNI's NewGlobalRef()
       "BsfUnloadJava",             // unload the JVM, if loaded by Rexx, calling JNI's DeleteGlobalRef()
       "BsfDropFuncs",              // unregister all of these functions
       "BsfVersion",                // returns the version number in the form "MajorNumber"."YYYYMMDD"
       "BsfQueryAllFunctions",       // fills in a stem with the names of all functions;
                                    // this *MUST* be the last listed function; it does not get registered;
                                    // as this was done already on the Rexx side;
       "BsfQueryRegisteredFunctions",// fills in a stem with the names of all functions;
                                    // this *MUST* be the last listed function; it does not get registered;
                                    // as this was done already on the Rexx side;
       "BsfLoadFuncs"               // already loaded (registered) by the Rexx programmer
   };

   // 2003-01-20, pointer to functions (for registering: same sequence as above!)
PFN  PApiFncTable[] =
   {
       (PFN) BSF,                       // interface for Rexx to Java
       (PFN) BsfInvokedBy,              // returns "0" (no invocation, hence no Java present),
                                        //   "1" (bsf4rexx loaded via Java), "2" (bsf4rexx loaded via Rexx)
       (PFN) BsfLoadJava,               // load the JVM by creating a new instance of BSFManager, if loaded by Rexx,
       (PFN) BsfUnloadJava,             // unload the JVM, if loaded by Rexx, calling JNI's DeleteGlobalRef()
       (PFN) BsfDropFuncs,              // unregister all of these functions
       (PFN) BsfVersion,                // returns the version number of this package
       (PFN) BsfQueryAllFunctions,       // fills in a stem with the names of the BSF functions
       (PFN) BsfQueryRegisteredFunctions,// fills in a stem with the names of the BSF functions

       (PFN) BsfLoadFuncs               // already loaded (registered) by the Rexx programmer
   };


   // 2003-01-26, array of functions to be loaded, if BSF4Rexx is invoked/loaded by Java
PSZ  ApiFncTable4Java[] =     // table of functions to be registered with Rexx
   {
       "BSF",                           // interface for Rexx to Java
       "BsfInvokedBy",                  // allows querying for invocation type of this DLL
       "BsfVersion",                    // returns the version number in the form "MajorNumber"."YYYYMMDD"
       "BsfQueryAllFunctions",           // fills in a stem with the names of all functions
       "BsfQueryRegisteredFunctions"     // fills in a stem with the names of all functions
   };

   // 2003-01-26, pointer to functions (for registering: same sequence as above!)
PFN  PApiFncTable4Java[] =
   {
       (PFN) BSF,                       // interface for Rexx to Java
       (PFN) BsfInvokedBy,              // returns "0" (no invocation, hence no Java present), //   "1" (bsf4rexx loaded via Java), "2" (bsf4rexx loaded via Rexx)
       (PFN) BsfVersion,                // unregister all of these functions
       (PFN) BsfQueryAllFunctions,       // fills in a stem with the names of the BSF functions
       (PFN) BsfQueryRegisteredFunctions // fills in a stem with the names of the BSF functions
   };




    // short of a make_upper macro/function, this is a little hack, ---rgf, 2003-01-27
inline void make_upper (char * string)
{
    for (; *string; string++)   // loop over characters
    {
        *string=toupper(*string);   // translate characters to uppercase
    }
}

/*-----------------------------------------------------------------------------
 * Create a REXX variable of the specified name and bind the value to it.
 *
 * [Thanks to Mark Hessling (2003-01-27) !!]
 *
 *----------------------------------------------------------------------------*/
int SetRexxVariable
   ( char *name, int namelen, char *value, int valuelen )
{
   ULONG  rc=0L;
   SHVBLOCK       shv;

   // already done elsewhere ( void )make_upper( name ); /* upper-case the variable name */
   shv.shvnext = ( SHVBLOCK* )NULL;
   shv.shvcode = RXSHV_SET;
   MAKERXSTRING( shv.shvname, name, ( ULONG )namelen );
   MAKERXSTRING( shv.shvvalue, value, ( ULONG )valuelen );
   shv.shvnamelen = shv.shvname.strlength;
   shv.shvvaluelen = shv.shvvalue.strlength;
   // assert( shv.shvname.strptr );
   // assert( shv.shvvalue.strptr );
   rc = RexxVariablePool( &shv );
   if ( rc == RXSHV_OK
   ||   rc == RXSHV_NEWV )
      return( 0 );
   else
      return( 1 );
}


/*********************************************************************/
/* Queries how BSF4Rexx was invoked and whether JVM is present       */
/*    returns: 0 ... no JVM present                                  */
/*             1 ... JVM present, bsf4rexx invoked by Java           */
/*             2 ... JVM present, bsf4rexx invoked by Rexx as an     */
/*                   external function package                       */
/*                                                                   */
/* ---rgf, 2003-01-21                                                */
/*********************************************************************/
APIRET APIENTRY BsfInvokedBy(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{
// fprintf(stderr, "BsfInvokedBy(): arrived\n"); zeigeGlobs();    // show values for global variables

    if (globObj == NULL)                // not invoked yet or JVM got unloaded in the meantime
    {
        strcpy(retstr->strptr, "0");
    }
    else if (globJvm == NULL)           // invoked by Java via BSF (bsf4rexx)
    {
        strcpy(retstr->strptr, "1");
    }
    else                                // invoked by Rexx (bsf4rexx), JVM is loaded
    {
        strcpy(retstr->strptr, "2");
    }
    retstr->strlength = strlen(retstr->strptr); // set proper string length

// fprintf(stderr, "BsfInvokedBy(): about to leave, \n");

  return VALID_ROUTINE;
}



/*********************************************************************/
/* load Java,                                                        */
/*    arg(1): optional classpath, if not given CLASSPATH will be used*/
/*            (BEWARE: CLASSPATH path *must* not end in a slash !!)  */
/*                                                                   */
/*    arg(2): optional, "P"repend, "A"ppend (default) classpath      */
/*            to the Java known classpath                            */
/*            (this is for Java 1.1.8 *only*, will have no effect    */
/*            if used on later Java versions)                        */
/*                                                                   */
/* returns:                                                          */
/*            0 ... JVM got loaded                                   */
/*            1 ... JVM got loaded in earlier runs, merely made      */
/*                  it available again                               */
/*              ... else, whatever JNI returned                      */
/*         -999 ... if JVM already loaded                            */
/*         -998 ... first argument not Append nor Prepend            */
/*         -997 ... argument #1 exceeds 1023 bytes                   */
/*         -996 ... length of Java-classpath plus the addition of    */
/*                  argument #1 (including PATH separator)           */
/*                  exceeds 1023 bytes                               */
/*         -995 ... could not create JVM                             */
/*         -994 ... init error: could not find 'Java4Rexx.class'     */
/*         -993 ... init error: could not find static method         */
/*                  'Java4Rexx.createInterface4Rexx()'               */
/*                                                                   */
/* ---rgf, 2003-01-15,21; switched args on 2003-02-23                */
/*********************************************************************/
APIRET APIENTRY BsfLoadJava(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{
#ifdef JNI_H
        // JNI 1.1 only ("JNI_H" is defined in 'jni.h' of version 1.1.x only!)
    JDK1_1InitArgs  vm_args;
#else
        // JNI 1.2 and above
    JavaVMInitArgs vm_args;
#endif

    jint            res;
    jclass          cls;
    jmethodID       mid;
    jstring         str;

    char classpath[1048]="";            // reserve space for classpath value
    int bAppend = 1;                    // default: append classpath to existing one
    char add_classpath_chunk[1024]="";  // chunk to add to classpath
    char *vmStartupCP="";             // VM startup classpath (for 1.1)

// fprintf(stderr, "BsfLoadJava(): just arrived\n"); zeigeGlobs();    // show values for global variables

        // JVM already loaded?
    if (globJvm!=NULL || globEnv!=NULL) {
        strcpy(retstr->strptr, "-999");
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

        // According to Sun's JNI tutorial, JNI 1.1 cannot unlaod the JVM, hence
        // this test for re-using the loaded JVM (otherwise a runtime error occurs)
    if (oldGlobJvm != NULL) // was JVM loaded already in this run, if so restore?
    {
            // restore old pointers/handles
        globJvm  = oldGlobJvm ;
        globORef = oldGlobORef;
        globEnv  = oldGlobEnv ;
        globObj  = oldGlobObj ;
        globMid  = oldGlobMid ;

        strcpy(retstr->strptr, "1");    // return "1" to indicate a re-use did take place
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return VALID_ROUTINE;
    }

#ifdef JNI_H
        // JNI 1.1 only ("JNI_H" is defined in 'jni.h' of version 1.1.x only!)
    vm_args.version = JNI_VERSION_1_1 ; // 0x00010001;
#else
        // JNI 1.2 and above
    vm_args.version = JNI_VERSION_1_2 ; // 0x00010002;
#endif

    jint resVer=JNI_GetDefaultJavaVMInitArgs ( &vm_args );  // get Java-arg structure from JNI

#ifdef JNI_H
        // JNI 1.1 only ("JNI_H" is defined in 'jni.h' of version 1.1.x only!)
    vm_args.version = JNI_VERSION_1_1 ; // 0x00010001;
    vmStartupCP=(char *) vm_args.classpath;      // get classpath
#endif

    char *cp_env=getenv("CLASSPATH");   // get value of environment variable CLASSPATH (NULL, if not set)

        // deal with received arguments
        // is a classpath argument supplied ?
    if ( numargs>0 )      // arguments are supplied, deal with them
    {
        if (args[0].strptr==NULL || args[0].strlength==0)
        {
                strcpy(add_classpath_chunk, ( cp_env==NULL ? "" : cp_env) );   // copy CLASSPATH value to var
        }
        else        // o.k. classpath argument supplied, check it
        {
            if ( args[0].strlength > 1023)     // test for hardcoded limits
            {
                fprintf(stderr, "BsfLoadJava(): argument # 1 exceeds 1023 bytes, aborting!\n");
                strcpy(retstr->strptr, "-997");
                retstr->strlength = strlen(retstr->strptr); // set proper string length
                return INVALID_ROUTINE;
            }
            strcpy(add_classpath_chunk, args[0].strptr);        // copy argument value to var
        }



#ifdef JNI_H
            // argument # 2 (Append, Prepend) given?, only necessary for Java 1.1
        if (numargs>1 && args[1].strptr!=NULL && args[1].strlength!=0)     // argument is there
        {
            char c = args[1].strptr[0]; // get character
            c=toupper(c);               // translate to uppercase

            switch (toupper(c))         // check upon right usage
            {
                case    'A': break;     // already set

                case    'P': bAppend=0; // prepend instead
                             break;
                default :
                    strcpy(retstr->strptr, "-998");
                    retstr->strlength = strlen(retstr->strptr); // set proper string length
                    return INVALID_ROUTINE;
            }
        }

            // check total length of new classpath
        // JNI 1.1 only ("JNI_H" is defined in 'jni.h' of version 1.1.x only!)
        if ( ( strlen(vmStartupCP) + strlen(add_classpath_chunk)) > 1022) {
            fprintf(stderr, "BsfLoadJava(): new assembled classpath exceeds 1023 bytes, aborting!\n");
            strcpy(retstr->strptr, "-996");
            retstr->strlength = strlen(retstr->strptr); // set proper string length
            return INVALID_ROUTINE;
        }
#endif

    }
    else        // no arguments, hence query environment for value for CLASSPATH
    {
        strcpy(add_classpath_chunk, ( cp_env==NULL ? "" : cp_env) );   // copy CLASSPATH value to var
    }


    /*  create the JVM */

#ifdef JNI_H
        // JNI 1.1 only ("JNI_H" is defined in 'jni.h' of version 1.1.x only!)
    bAppend ? sprintf(classpath, "%s%c%s", vmStartupCP,         PATH_SEPARATOR, add_classpath_chunk) :
              sprintf(classpath, "%s%c%s", add_classpath_chunk, PATH_SEPARATOR, vmStartupCP)   ;
    vm_args.classpath=classpath;        // assign new value
    res = JNI_CreateJavaVM( &globJvm, &globEnv, &vm_args);

#else
        // JNI 1.2 and above
    JavaVMOption options[1];                // define an array for Java startup options
    sprintf(classpath, "-Djava.class.path=%s", add_classpath_chunk);           /* disable JIT */
    options[0].optionString = classpath;

    vm_args.version = JNI_VERSION_1_2;      // set version
    vm_args.options = options;              // set options-array (containing classpath)
    vm_args.nOptions = 1;                   // indicate how many entries in array
    vm_args.ignoreUnrecognized = FALSE;     // report unrecognized values !
    res = JNI_CreateJavaVM( &globJvm, (void **) &globEnv, &vm_args);
#endif

    if (res<0) {
        fprintf(stderr, "BsfLoadJava(): Can't create Java VM, rc=%d\n", res);
        strcpy(retstr->strptr, "-995");
        // sprintf(retstr->strptr, "%d", res);         // copy return code
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    /* Find the class "Java4Rexx" */
    cls = (*globEnv).FindClass(JAVA_4_REXX);    // or: cls = globEnv->FindClass(JAVA_4_REXX);
    if (cls == 0) {
        fprintf(stderr, "BsfLoadJava(): initialization error, can't find class: \n\t'%s'\n", JAVA_4_REXX);
        strcpy(retstr->strptr, "-994");
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    /* get static Method "createInterface4Rexx()" */
     mid = globEnv->GetStaticMethodID(cls, "createInterface4Rexx", "()V");
    if (mid == 0) {
        fprintf(stderr, "BsfLoadJava(): initialization error, can't find method: \n\t'Java4Rexx.createInterface4Rexx()'\n");
        strcpy(retstr->strptr, "-993");
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    /* call the static method, which in turn will initialize BSF4Rexx by creating
       a BSFManager and having it load the "Rexx" (support) */
    strcpy(retstr->strptr, "0");
    retstr->strlength = strlen(retstr->strptr); // set proper string length

    globEnv->CallStaticVoidMethod(cls, mid, NULL);      // invoke static method

// fprintf(stderr, "BsfLoadJava(): about to leave, \n"); zeigeGlobs();    // show values for global variables
  return VALID_ROUTINE;
}


// do the "gory" work
long RgfUnloadJava()
{
    long ret=1l;    // default return code, indicate that no real unloading took place (JNI 1.1 restriction)
        // save old pointers/handles
    oldGlobJvm = globJvm  ;
    oldGlobORef= globORef ;
    oldGlobEnv = globEnv  ;
    oldGlobObj = globObj  ;
    oldGlobMid = globMid  ;

/* unloading/destroying of the JVM will not really unload the JVM, hence keep it for
   possible later re-usage */

/* this would be the code to destroy/unload the JVM

        // o.k. now reset everything to initial state
    if (globORef != NULL) {
        globEnv->DeleteGlobalRef(globORef); // delete global reference
        globORef=NULL;      // don't use anymore, set to NULL
    }

     long  ret=globJvm->DestroyJavaVM ( );    // destroy JVM
*/

        // reset global variables
    globJvm  = NULL;
    globORef = NULL;
    globEnv  = NULL;
    globObj  = NULL;
    globMid  = NULL;

    return ret;
}


/*********************************************************************/
/* unload Java, according to tutorial: does not really work on       */
/*              Java 1.1, will always return an error                */
/*                                                                   */
/* returns:                                                          */
/*         -1   ... indicate everything o.k., but JVM cached (JNI1.1)*/
/*         -999 ... JVM was not loaded by Rexx, aborting             */
/*         -998 ... no JVM loaded at present, aborting               */
/*                                                                   */
/* ---rgf, 2003-01-15,21, 26                                         */
/*********************************************************************/
APIRET APIENTRY BsfUnloadJava(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{
    JDK1_1InitArgs  vm_args;
    jint            res;
    jclass          cls;
    jmethodID       mid;
    jstring         str;

// fprintf(stderr, "BsfUnloadJava(): just arrived, \n"); zeigeGlobs();    // show values for global variables

    if (globJvm == NULL) // JVM was not loaded by Rexx, hence show error
    {
        fprintf(stderr, "BSFUnloadJava(): JVM was not started by Rexx, aborting!\n");
        strcpy(retstr->strptr, "-999"); // indicate no JVM loaded
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    if (globEnv == NULL) // JVM is not loaded, error!
    {
        fprintf(stderr, "BSFUnloadJava(): no JVM to unload, aborting!\n");
        strcpy(retstr->strptr, "-998");             // indicate JVM is not loaded
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    long ret=RgfUnloadJava();   // do the unloading, ignore anything for the moment (as long as JNI 1.1 is used)

    //  sprintf(retstr->strptr, "%d", ret);     // result: <> 0 failed
    strcpy(retstr->strptr, "1");     // always return "1" to indicate that JVM was not really unloaded
    retstr->strlength = strlen(retstr->strptr); // set proper string length

// fprintf(stderr, "BsfUnloadJava(): about to leave... \n"); zeigeGlobs();    // show values for global variables

    return VALID_ROUTINE;
}




/*********************************************************************/
/* load (register) all BSF-functions                                 */
/*                                                                   */
/* returns:                                                          */
/*                                                                   */
/*            0 ... registering went o.k.                            */
/*         -999 ... 'BSF' already loaded, hence abort                */
/*         -998 ... at least one registered function caused a        */
/*                  non-zero return code, hence abort                */
/*                                                                   */
/*********************************************************************/
APIRET APIENTRY BsfLoadFuncs(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{
  INT    entries;                      /* Num of entries             */
  INT    j;                            /* Counter                    */

//fprintf(stderr, "BsfLoadFuncs(): just arrived.\n");

  if (RexxQueryFunction("BSF") != RXFUNC_NOTREG )     // "BSF" already registered!
  {
      fprintf(stderr, "BsfLoadFuncs(): 'BSF' already registered, aborting!\n");
      strcpy(retstr->strptr, "-999"); // indicate BSF already loaded (probably by Java)
      retstr->strlength = strlen(retstr->strptr); // set proper string length
      return INVALID_ROUTINE;
  }

  entries = sizeof(ApiFncTable)/sizeof(PSZ);    // table with function names to be registered
  entries--;        // make sure BsfLoadFuncs() is not registered (already happened at the Rexx side)

//fprintf(stderr, "\t... registering %d functions with Rexx ...\n", entries);

  long res=0l, resTmp=0l;                  // initialize to 0l
  for (j = 0; j < entries; j++)
  {
      // 2003-01-20, making functions process-local by using RexxRegisterFunctionExe() instead of ...Dll()
      resTmp = RexxRegisterFunctionExe(ApiFncTable[j], PApiFncTable[j]);

// fprintf(stderr, "===> BsfLoadFuncs(): j=%d, resTmp=%d, '%s'\n", j, resTmp, ApiFncTable[j]);

      if (resTmp != 0l)     // something went wrong with registering this function
      {
          fprintf(stderr,"BsfLoadFuncs(): error '%d' while registering function '%s'\n", resTmp, ApiFncTable[j]);
      }
      res+=resTmp;      // just add them up
  }


  if (res != 0l)          // at least one of the functions could not be registerd, abort
  {
      strcpy(retstr->strptr, "-998"); // indicate that registering caused at least one problem
      retstr->strlength = strlen(retstr->strptr); // set proper string length
      return INVALID_ROUTINE;
  }

  strcpy(retstr->strptr, "0");     // registering went o.k.
  retstr->strlength = strlen(retstr->strptr); // set proper string length

  return VALID_ROUTINE;
}


/*********************************************************************/
/* unregisters all BSF-functions                                     */
/*                                                                   */
/* returns:                                                          */
/*                                                                   */
/*            0 ... deregistering went o.k.                          */
/*         -998 ... at least one deregistering function caused a     */
/*                  non-zero return code, hence abort                */
/*                                                                   */
/*********************************************************************/

APIRET APIENTRY BsfDropFuncs(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{

// fprintf(stderr, "BsfDropFuncs(): just arrived\n"); zeigeGlobs();    // show values for global variables

  INT    entries;                      /* Num of entries             */
  INT    j;                            /* Counter                    */
  long res=0l, resTmp=0l;               // initialize to 0l

  entries = sizeof(ApiFncTable)/sizeof(PSZ);

  long ret=RgfUnloadJava();     // unload the JVM (pseudo unload due to JNI 1.1 restrictions)

  for (j = 0; j < entries; j++)
  {
// fprintf(stderr, "===> BsfDropFuncs(), 1) RexxQueryFunc('%s')=%d\n", ApiFncTable[j], RexxQueryFunction(ApiFncTable[j]));
      resTmp=RexxDeregisterFunction(ApiFncTable[j]);
// fprintf(stderr, "===> BsfDropFuncs(), 2) resTmp=%d, function: '%s'\n", resTmp, ApiFncTable[j]);

      if (resTmp != 0l)                 // something went wrong
      {
          fprintf(stderr,"BsfDropFuncs(): error '%d' while deregistering function '%s'\n", resTmp, ApiFncTable[j]);
      }
      res+=resTmp;

            // hmm, Object Rexx returns "0", Regina returns "30" ("function not registered") ???
            // if registering thereafter, Regina returns "10" ("function already registered") !
  }

  if (res != 0l)          // at least one of the functions could not be deregistered, abort
  {
      strcpy(retstr->strptr, "-998"); // indicate something went wrong while deregistering
      retstr->strlength = strlen(retstr->strptr); // set proper string length
      return INVALID_ROUTINE;
  }

  strcpy(retstr->strptr, "0");                  // deregistering went o.k.
  retstr->strlength = strlen(retstr->strptr);   // set proper string length

  return VALID_ROUTINE;
}


/*********************************************************************/
/* returns the version number                                        */
/*                                                                   */
/* returns:                                                          */
/*                                                                   */
/*            x.YYYYMMDD "MajorNumber"."Date"                        */
/*                                                                   */
/*********************************************************************/

APIRET APIENTRY BsfVersion(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{

// fprintf(stderr, "BsfVersion(): just arrived\n");

  strcpy(retstr->strptr, BSF_VERSION);          // version number
  retstr->strlength = strlen(retstr->strptr);   // set proper string length

  return VALID_ROUTINE;
}


/*********************************************************************/
/* fills in a stem with the function names                           */
/*                                                                   */
/* returns:                                                          */
/*                                                                   */
/*            0 ... function executed o.k.                           */
/*         -999 ... supplied stem name too long, hence abort         */
/*                                                                   */
/*********************************************************************/

APIRET APIENTRY BsfQueryAllFunctions(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{

// fprintf(stderr, "BsfQueryAllFunctions(): just arrived\n");

    char varname[255], value[10];
    char stem_name[255];
    int entries = sizeof(ApiFncTable)/sizeof(PSZ), rc;  // table with function names to be registered
    int j=0; // --rgf, 2003-04-03/2003-04-14: keeping gcc happy


    if (numargs==0 || args[0].strlength==0)     // no stem-name given, use "BSFLISTFUNCTIONS." as default
    {
        strcpy( stem_name, "BsfQueryAllFunctions.");  // default name for this stem
    }

    else if (args[0].strlength>244)             // variable name too long 244+dot(1char)+max_digits(=10)+string_end_char(x00)
    {
        fprintf(stderr, "BsfQueryAllFunctions(): supplied stem name ('%s') too long ('%d' characters exceed the maximum of 244 characters), aborting!\n", args[0].strptr, args[0].strlength);
        strcpy(retstr->strptr, "-999");         // indicate stem-name too long
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    else
    {
        strcpy( stem_name, args[0].strptr);       // use whatever the user supplied
    }

    make_upper(stem_name);      // uppercase the stem-name

    for (j = 0; j < entries; j++)
    {
            (void)sprintf(varname, "%s%u", stem_name, j+1);
            rc = SetRexxVariable( varname, strlen(varname), ApiFncTable[j] , strlen(ApiFncTable[j]) );
    }

    sprintf(value, "%u", j);                    // give number of elements in stem array
    (void)sprintf(varname, "%s0", stem_name);
    rc = SetRexxVariable( varname, strlen(varname), value, strlen(value));


  strcpy(retstr->strptr, "0");                  // return value
  retstr->strlength = strlen(retstr->strptr);   // set proper string length

  return VALID_ROUTINE;
}



/*********************************************************************/
/* fills in a stem with the function names                           */
/*                                                                   */
/* returns:                                                          */
/*                                                                   */
/*            0 ... function executed o.k.                           */
/*         -999 ... supplied stem name too long, hence abort         */
/*                                                                   */
/*********************************************************************/

APIRET APIENTRY BsfQueryRegisteredFunctions(
  PSZ       name,                      /* Function name              */
  LONG      numargs,                   /* Number of arguments        */
  RXSTRING  args[],                    /* Argument array             */
  PSZ       queuename,                 /* Current queue              */
  PRXSTRING retstr )                   /* Return RXSTRING            */
{

// fprintf(stderr, "BsfQueryRegisteredFunctions(): just arrived\n");

    char varname[255], value[10];
    char stem_name[255];
    int entries, rc;

    if (numargs==0 || args[0].strlength==0)     // no stem-name given, use "BSFLISTFUNCTIONS." as default
    {
        strcpy( stem_name, "BsfQueryRegisteredFunctions.");  // default name for this stem
    }

    else if (args[0].strlength>244)             // variable name too long 244+dot(1char)+max_digits(=10)+string_end_char(x00)
    {
        fprintf(stderr, "BsfQueryRegisteredFunctions(): supplied stem name ('%s') too long ('%d' characters exceed the maximum of 244 characters), aborting!\n", args[0].strptr, args[0].strlength);
        strcpy(retstr->strptr, "-999");         // indicate stem-name too long
        retstr->strlength = strlen(retstr->strptr); // set proper string length
        return INVALID_ROUTINE;
    }

    else
    {
        strcpy( stem_name, args[0].strptr);       // use whatever the user supplied
    }

    make_upper(stem_name);      // uppercase the stem-name

    entries=sizeof(ApiFncTable)/sizeof(PSZ);
    int i=0, j=0;

    for (i=0; i<entries; i++) {             // loop over all functions
        if (! RexxQueryFunction(ApiFncTable[i]))    // is the function registered at the moment?
        {
            (void)sprintf(varname, "%s%u", stem_name, ++j);
            rc = SetRexxVariable( varname, strlen(varname), ApiFncTable[i] , strlen(ApiFncTable[i]) );
        }
    }

    sprintf(value, "%u", j);                    // give number of elements in stem array
    (void)sprintf(varname, "%s0", stem_name);
    rc = SetRexxVariable( varname, strlen(varname), value, strlen(value));

  strcpy(retstr->strptr, "0");                  // return value
  retstr->strlength = strlen(retstr->strptr);   // set proper string length

  return VALID_ROUTINE;
}




/*
 * External Rexx function "BSF". This function allows Rexx programs to call into Java.
 */
APIRET APIENTRY
BSF  (  PSZ name,
        ULONG argc,
        PRXSTRING argv,
        PSZ queuename,
        PRXSTRING retstr)
     {


// fprintf(stderr, "BSF(): just arrived.\n"); zeigeGlobs();    // show values for global variables

        //convert arguments array to Java format
        jobjectArray arr = (*globEnv).NewObjectArray (
                                      argc,
                                    (*globEnv).FindClass("java/lang/String"),
                                    (*globEnv).NewStringUTF("")
                                                     );

        // RXSTRINGs passed from Rexx have an additional \0 appended according Object Rexx 2.1 docs
        for(ULONG i=0; i<argc; i++)
        {
            if (argv[i].strptr==NULL)   // NULL (argument missing) ?
            {
               (*globEnv).SetObjectArrayElement(arr, i, NULL);
            }
            else                        // argument as a string in hand
            {
               (*globEnv).SetObjectArrayElement(arr, i,
                                               (*globEnv).NewStringUTF(argv[i].strptr));
            }
        }

        // call Java method and create Rexx return value
        jstring ReturnValue = (jstring)(*globEnv).CallObjectMethod(globObj, globMid, arr);

            // checking return value from Java
        if (ReturnValue!=NULL)          // value returned from Java?
        {
                const char *str = globEnv->GetStringUTFChars(ReturnValue, 0);
                int         len = strlen(str);

                // o.k. not enough space available, free Rexx-allocated memory, re-allocate new chunk
//                if (len>255)
                if (len>=RXAUTOBUFLEN)  // if not enough space available, get a larger chunk and use it
                {
// ---rgf, do not free memory from interpreter!                   RgfFreeMemory ( retstr->strptr );                    // free buffer

                   retstr->strptr = (char *) RgfAllocateMemory(len+1) ; // allocate new buffer
                }

                strcpy(retstr->strptr, str);
                retstr->strlength = len;
                globEnv->ReleaseStringUTFChars(ReturnValue, str);
        }
        else    // o.k. Java returned null
        {
/* rgf, 2003-04-14, never free memory allocated by interpreter!
            // rgf, 2001-05
            if (retstr->strptr != NULL)
            {
                 RgfFreeMemory ( retstr->strptr );
            }
*/

            retstr->strptr    = NULL;   // set string pointer to NULL (as if no value was returned)
            retstr->strlength = 0;      // indicate no chars available
        }

                // pending exception ?
        if ( globEnv->ExceptionOccurred() != NULL ) // this works on JRE 1.1.8 ! ---rgf, 2001-05-25
        {
             globEnv->ExceptionDescribe();      // print the stack
             globEnv->ExceptionClear();         // clear exception in the JVM
             return INVALID_ROUTINE;    // interrupt Rexx with "Incorrect call to routine!"
        }

        return VALID_ROUTINE;
     }



/*
 * This will be called from Java at initialization time and makes sure
 * that the external Rexx functions named &quot;BsfInvokedBy()&quot; and
 * &quot;BSF()&quot; will get
 * registered. This way, Rexx programs being invoked from the Java side
 * have &quot;BSF&quot; available already (no need to register this function by
 * called Rexx program). Changed to "jint" on 2001-05-25.
 */
JNIEXPORT jint JNICALL

#ifdef APACHE
Java_org_apache_bsf_engines_rexx_RexxAndJava_jniRegisterBSF (JNIEnv *env, jobject obj)
#else
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRegisterBSF (JNIEnv *env, jobject obj)
#endif

     {
            // 2003-01-26
         int entries = sizeof(ApiFncTable4Java)/sizeof(PSZ);    // table with function names to be registered

            // register
         long res=0l, resTmp=0l;                  // initialize to 0l
         for (int j = 0; j < entries; j++)
         {
             // 2003-01-20, making functions process-local by using RexxRegisterFunctionExe() instead of ...Dll()
             resTmp = RexxRegisterFunctionExe(ApiFncTable4Java[j], PApiFncTable4Java[j]);
// fprintf(stderr, "..._jniRegisterBSF(): j=%d, resTmp=%d, '%s'\n", j, resTmp, ApiFncTable4Java[j]);
         }

         return (jint) 1;   // always return true
     }


/*
 * This may be called from Java, pertains to Object Rexx only (as of 2003-01-04);
 * should wait until all cleanup within Object Rexx has been done. ---rgf, 2002-12-07, 2003-01-04
 */
JNIEXPORT jint JNICALL

#ifdef APACHE
Java_org_apache_bsf_engines_rexx_RexxAndJava_jniRexxWaitForTermination
#else
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxWaitForTermination
#endif
  (JNIEnv *env, jobject obj)
     {
// fprintf(stderr, "..._jniRexxWaitForTermination(), just arrived. ...\n");

#if defined (USE_OREXX) || defined( USE_REXXTRANS )
// fprintf(stderr, "..._jniRexxWaitForTermination(), just about to call...\n");
        RexxWaitForTermination();       // wait for Object Rexx to re-synch ... (2003-02-25, ---rgf: doesn't really have any effect
#endif
// fprintf(stderr, "..._jniRexxWaitForTermination() about to return...\n");

        return (jint) 1;               // just return something

     }
  ;


/*
 * This is a pass-thru for RexxDidRexxTerminate(), returns 0 if Object Rexx still runs,
 * 1 if Object Rexx has terminated, ---rgf, 2002-12-13
 */
JNIEXPORT jint JNICALL

#ifdef APACHE
Java_org_apache_bsf_engines_rexx_RexxAndJava_jniRexxDidRexxTerminate
#else
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxDidRexxTerminate
#endif

  (JNIEnv *env, jobject obj)
     {
//fprintf(stderr, "..._jniRexxDidRexxTerminate() ...\n");

#if defined (USE_OREXX) || defined( USE_REXXTRANS )
        return (jint) RexxDidRexxTerminate();       // ask Object Rexx, if it is done

#else
        return (jint) 1;        // default for non-Object Rexx is the integer value '1' for 'true'
#endif
     }
  ;






/*

 * Class:     com_ibm_bsf_engines_rexx_RexxAndJava

 * Method:    jniInitialize4Rexx

 * Signature: ()I

 *
 * This is the target from the Java side. Will be called upon initializing
 * the JVM from the Rexx side.
 *
 */
#ifdef APACHE
JNIEXPORT jint JNICALL Java_org_apache_bsf_engines_rexx_RexxAndJava_jniInitialize4Rexx
#else
JNIEXPORT jint JNICALL Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniInitialize4Rexx
#endif

  (JNIEnv *env, jobject obj)
{
// fprintf(stderr, "..._jniInitialize4Rexx(): just arrived\n"); zeigeGlobs();    // show values for global variables

    globORef = obj = (*env).NewGlobalRef(obj);  // pin down the Java object which called this function
    set_global_vars(env, obj);    // save handles "globally"

    return (jint) 1;              // just return something which can be thought of being true
}

/*
 * This will be called from Java to invoke the Rexx interpreter. "source" contains
 * the Rexx program, "array" the arguments meant for the Rexx program.
 */
#ifdef APACHE
JNIEXPORT jstring JNICALL Java_org_apache_bsf_engines_rexx_RexxAndJava_jniRexxStart
#else
JNIEXPORT jstring JNICALL Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxStart
#endif
        ( JNIEnv *env,
          jobject         obj,
          jstring         source,
          jobjectArray    array)
     {
         // set global variables for callbacks to Java (via BSF())
// fprintf(stderr, "..._jniRexxStart(): just arrived, \n"); zeigeGlobs();    // show values for global variables

         set_global_vars(env, obj); // save handles "globally"

         // make a copy of the Java string containing the Rexx script to be executed
         const char      *str = env->GetStringUTFChars(source, 0);
         char      *sourceStr = (char *) RgfAllocateMemory(strlen(str)+1) ;

         strcpy(sourceStr, str);
         env->ReleaseStringUTFChars(source, str);

         // create argument array for Rexx
         PRXSTRING argList = NULL;       // pointer to array of RXSTRINGS
         jsize   arrNum = 0;
         long    numArgs= 0l;

         if (array != NULL)              // an Object array available?
         {
             arrNum    =  env->GetArrayLength(array);
             numArgs   = (long) arrNum;
         }

         if (arrNum>0)
         {
                // allocate enough memory for argument array of RXSTRINGs
            argList = (PRXSTRING) RgfAllocateMemory( sizeof(RXSTRING) * arrNum );

            for (int i = 0; i < arrNum; ++i)
            {
               if ((env->GetObjectArrayElement(array, i))==NULL)/* no argument*/
               {
                  argList[i].strptr   =NULL;
                  argList[i].strlength=0;
               }
               else        /* a string argument in hand, give it to Rexx*/
               {
                  jstring        javaStr = (jstring) env->GetObjectArrayElement(array, i);
                  const char     *strArg = env->GetStringUTFChars(javaStr, 0);
                  int                len = strlen(strArg);
                  char        *tmpString =  (char *) RgfAllocateMemory(len+1) ;

                  strcpy(tmpString, strArg);
                  env->ReleaseStringUTFChars(javaStr, strArg);

                  MAKERXSTRING( argList[i], tmpString, len);/* save arg for Rexx*/
               }
            }
         }

                // Rexx program to execute
         RXSTRING        RXsourceStr[2];
         MAKERXSTRING   (RXsourceStr[0], sourceStr, strlen(sourceStr));
         PRXSTRING       Instore = &RXsourceStr[0];
         MAKERXSTRING   (RXsourceStr[1], NULL, 0);

         SHORT           ReturnCode;
         LONG            ReturnValue;

                // let Rexx allocate buffer, if needed for a result
         RXSTRING        Result;
         MAKERXSTRING   (Result, NULL, 0);

         PSZ             ProgramName = DLLNAME ;

// printf("==> JNI-jniRexxStart: about to invoke Rexx via RexxStart(). \n");
         //start Rexx
         ReturnValue = RexxStart(
                         arrNum,         // LONG         ArgCount
                         argList,        // PRXSTRING    ArgList
                         ProgramName,    // PSZ          ProgramName
                         Instore,        // PRXSTRING    Instore
                         NULL,           // PSZ          EnvName
                         RXSUBROUTINE,   // LONG         CallType
                         NULL,           // PRXSYSEXIT   Exits
                         &ReturnCode,    // PUSHORT      ReturnCode
                         &Result         // PRXSTRING    Result
                         );

// fprintf(stderr, "..._jniRexxStart(): after invoking Rexx, \n"); zeigeGlobs();    // show values for global variables

/* ---rgf, 2003-04-14, never free memory allocated by interpreter!
                // freeing memory for Rexx tokenized image
        RgfFreeMemory ( RXsourceStr[1].strptr );// free tokenized image of it
*/

        if (arrNum>0)   // free allocated memory, which was needed for Rexx arguments
        {
           for (int i = 0; i < arrNum; ++i) // loop over Rexx args
           {
              RgfFreeMemory ( argList[i].strptr );  // free argument array RXSTRING buffers
           }
           RgfFreeMemory ( argList );               // free argument array RXSTRING buffer itself
        }

        RgfFreeMemory(sourceStr);   // ---rgf, 2003-04-14, free memory reserved for the Rexx source code

        if (ReturnValue != 0l)     // error occurred, return null
        {
/* ---rgf, 2003-04-14, never free memory allocated by Rexx!
           RgfFreeMemory ( Result.strptr );     // free memory allocated from Rexx
*/
           return NULL;
        }

         // --- get result and return it to Java
        jstring jResult=NULL;           // define Java String to return

        if ( Result.strptr != NULL )    // a result was returned from Rexx
        {
             if (Result.strlength==0)   // empty string?
             {
                 jResult=env->NewStringUTF("");
             }
             else                       // valid non-empty string
             {
                 jResult=env->NewStringUTF( Result.strptr );
             }

/* ---rgf, 2003-04-14, never free memory allocated by Rexx!
RgfFreeMemory ( Result.strptr ); // free memory where Rexx placed its result into
*/
        }

        return jResult;         // return a Java String
     }



#ifdef __cplusplus
}               // closing bracket
#endif


